@@ -1,38 +1,38 @@
-///////////////////////////////////////////////////////////////////////////
-//
-// Copyright (c) 2012-2018 DreamWorks Animation LLC
-//
-// All rights reserved. This software is distributed under the
-// Mozilla Public License 2.0 ( http://www.mozilla.org/MPL/2.0/ )
-//
-// Redistributions of source code must retain the above copyright
-// and license notice and the following restrictions and disclaimer.
-//
-// *     Neither the name of DreamWorks Animation nor the names of
-// its contributors may be used to endorse or promote products derived
-// from this software without specific prior written permission.
-//
-// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY INDIRECT, INCIDENTAL,
-// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-// IN NO EVENT SHALL THE COPYRIGHT HOLDERS' AND CONTRIBUTORS' AGGREGATE
-// LIABILITY FOR ALL CLAIMS REGARDLESS OF THEIR BASIS EXCEED US$250.00.
-//
-///////////////////////////////////////////////////////////////////////////
+// Copyright Contributors to the OpenVDB Project
+// SPDX-License-Identifier: Apache-2.0
 
 /// @file points/IndexFilter.h
 ///
 /// @author Dan Bailey
 ///
 /// @brief  Index filters primarily designed to be used with a FilterIndexIter.
+///
+/// Filters must adhere to the interface described in the example below:
+/// @code
+/// struct MyFilter
+/// {
+///     // Return true when the filter has been initialized for first use
+///     bool initialized() { return true; }
+///
+///     // Return index::ALL if all points are valid, index::NONE if no points are valid
+///     // and index::PARTIAL if some points are valid
+///     index::State state() { return index::PARTIAL; }
+///
+///     // Return index::ALL if all points in this leaf are valid, index::NONE if no points
+///     // in this leaf are valid and index::PARTIAL if some points in this leaf are valid
+///     template <typename LeafT>
+///     index::State state(const LeafT&) { return index::PARTIAL; }
+///
+///     // Resets the filter to refer to the specified leaf, all subsequent valid() calls
+///     // will be relative to this leaf until reset() is called with a different leaf.
+///     // Although a required method, many filters will provide an empty implementation if
+///     // there is no leaf-specific logic needed.
+///     template <typename LeafT> void reset(const LeafT&) { }
+///
+///     // Returns true if the filter is valid for the supplied iterator
+///     template <typename IterT> bool valid(const IterT&) { return true; }
+/// };
+/// @endcode
 
 #ifndef OPENVDB_POINTS_INDEX_FILTER_HAS_BEEN_INCLUDED
 #define OPENVDB_POINTS_INDEX_FILTER_HAS_BEEN_INCLUDED
@@ -42,16 +42,16 @@
 
 #include <openvdb/math/Transform.h>
 #include <openvdb/tools/Interpolation.h>
+#include <openvdb/util/Assert.h>
 
 #include "IndexIterator.h"
 #include "AttributeArray.h"
 #include "AttributeGroup.h"
 #include "AttributeSet.h"
 
-#include <boost/ptr_container/ptr_vector.hpp>
-
 #include <random> // std::mt19937
 #include <numeric> // std::iota
+#include <unordered_map>
 
 
 class TestIndexFilter;
@@ -64,6 +64,7 @@ namespace points {
 
 ////////////////////////////////////////
 
+/// @cond OPENVDB_DOCS_INTERNAL
 
 namespace index_filter_internal {
 
@@ -97,6 +98,39 @@ generateRandomSubset(const unsigned int seed, const IntType n, const IntType m)
 
 } // namespace index_filter_internal
 
+/// @endcond
+
+
+/// Index filtering on active / inactive state of host voxel
+template <bool On>
+class ValueMaskFilter
+{
+public:
+    static bool initialized() { return true; }
+    static index::State state() { return index::PARTIAL; }
+    template <typename LeafT>
+    static index::State state(const LeafT& leaf)
+    {
+        if (leaf.isDense())         return On ? index::ALL : index::NONE;
+        else if (leaf.isEmpty())    return On ? index::NONE : index::ALL;
+        return index::PARTIAL;
+    }
+
+    template <typename LeafT>
+    void reset(const LeafT&) { }
+
+    template <typename IterT>
+    bool valid(const IterT& iter) const
+    {
+        const bool valueOn = iter.isValueOn();
+        return On ? valueOn : !valueOn;
+    }
+};
+
+
+using ActiveFilter = ValueMaskFilter<true>;
+using InactiveFilter = ValueMaskFilter<false>;
+
 
 /// Index filtering on multiple group membership for inclusion and exclusion
 ///
@@ -142,22 +176,30 @@ public:
 
     inline bool initialized() const { return mInitialized; }
 
+    inline index::State state() const
+    {
+        return (mInclude.empty() && mExclude.empty()) ? index::ALL : index::PARTIAL;
+    }
+
+    template <typename LeafT>
+    static index::State state(const LeafT&) { return index::PARTIAL; }
+
     template <typename LeafT>
     void reset(const LeafT& leaf) {
         mIncludeHandles.clear();
         mExcludeHandles.clear();
-        for (const AttributeSet::Descriptor::GroupIndex& index : mInclude) {
-            mIncludeHandles.emplace_back(leaf.groupHandle(index));
+        for (const auto& i : mInclude) {
+            mIncludeHandles.emplace_back(leaf.groupHandle(i));
         }
-        for (const AttributeSet::Descriptor::GroupIndex& index : mExclude) {
-            mExcludeHandles.emplace_back(leaf.groupHandle(index));
+        for (const auto& i : mExclude) {
+            mExcludeHandles.emplace_back(leaf.groupHandle(i));
         }
         mInitialized = true;
     }
 
     template <typename IterT>
     bool valid(const IterT& iter) const {
-        assert(mInitialized);
+        OPENVDB_ASSERT(mInitialized);
         // accept no include filters as valid
         bool includeValid = mIncludeHandles.empty();
         for (const GroupHandle& handle : mIncludeHandles) {
@@ -188,7 +230,7 @@ class RandomLeafFilter
 {
 public:
     using SeedCountPair = std::pair<Index, Index>;
-    using LeafMap       = std::map<openvdb::Coord, SeedCountPair>;
+    using LeafMap       = std::unordered_map<openvdb::Coord, SeedCountPair>;
 
     RandomLeafFilter(   const PointDataTreeT& tree,
                         const Index64 targetPoints,
@@ -203,7 +245,7 @@ public:
         std::mt19937 generator(seed);
         std::uniform_int_distribution<unsigned int> dist(0, std::numeric_limits<unsigned int>::max() - 1);
 
-        Index32 leafCounter = 0;
+        Index64 leafCounter = 0;
         float totalPointsFloat = 0.0f;
         int totalPoints = 0;
         for (auto iter = tree.cbeginLeaf(); iter; ++iter) {
@@ -213,7 +255,7 @@ public:
                 mLeafMap[iter->origin()] = SeedCountPair(dist(generator), leafPoints);
                 break;
             }
-            totalPointsFloat += factor * iter->pointCount();
+            totalPointsFloat += factor * static_cast<float>(iter->pointCount());
             const auto leafPoints = static_cast<int>(math::Floor(totalPointsFloat));
             totalPointsFloat -= static_cast<float>(leafPoints);
             totalPoints += leafPoints;
@@ -226,6 +268,10 @@ public:
 
     inline bool initialized() const { return mNextIndex == -1; }
 
+    static index::State state() { return index::PARTIAL; }
+    template <typename LeafT>
+    static index::State state(const LeafT&) { return index::PARTIAL; }
+
     template <typename LeafT>
     void reset(const LeafT& leaf) {
         using index_filter_internal::generateRandomSubset;
@@ -297,15 +343,19 @@ public:
 
     inline bool initialized() const { return bool(mIdHandle); }
 
+    static index::State state() { return index::PARTIAL; }
+    template <typename LeafT>
+    static index::State state(const LeafT&) { return index::PARTIAL; }
+
     template <typename LeafT>
     void reset(const LeafT& leaf) {
-        assert(leaf.hasAttribute(mIndex));
+        OPENVDB_ASSERT(leaf.hasAttribute(mIndex));
         mIdHandle.reset(new Handle(leaf.constAttributeArray(mIndex)));
     }
 
     template <typename IterT>
     bool valid(const IterT& iter) const {
-        assert(mIdHandle);
+        OPENVDB_ASSERT(mIdHandle);
         const IntType id = mIdHandle->get(*iter);
         const unsigned int seed = mSeed + static_cast<unsigned int>(id);
         RandGenT generator(seed);
@@ -350,6 +400,10 @@ public:
 
     inline bool initialized() const { return bool(mPositionHandle); }
 
+    static index::State state() { return index::PARTIAL; }
+    template <typename LeafT>
+    static index::State state(const LeafT&) { return index::PARTIAL; }
+
     template <typename LeafT>
     void reset(const LeafT& leaf) {
         mPositionHandle.reset(new Handle(leaf.constAttributeArray("P")));
@@ -357,8 +411,8 @@ public:
 
     template <typename IterT>
     bool valid(const IterT& iter) const {
-        assert(mPositionHandle);
-        assert(iter);
+        OPENVDB_ASSERT(mPositionHandle);
+        OPENVDB_ASSERT(iter);
 
         const openvdb::Coord ijk = iter.getCoord();
         const openvdb::Vec3f voxelIndexSpace = ijk.asVec3d();
@@ -410,6 +464,13 @@ public:
 
     inline bool initialized() const { return bool(mPositionHandle); }
 
+    inline index::State state() const
+    {
+        return mBbox.empty() ? index::NONE : index::PARTIAL;
+    }
+    template <typename LeafT>
+    static index::State state(const LeafT&) { return index::PARTIAL; }
+
     template <typename LeafT>
     void reset(const LeafT& leaf) {
         mPositionHandle.reset(new Handle(leaf.constAttributeArray("P")));
@@ -417,7 +478,7 @@ public:
 
     template <typename IterT>
     bool valid(const IterT& iter) const {
-        assert(mPositionHandle);
+        OPENVDB_ASSERT(mPositionHandle);
 
         const openvdb::Coord ijk = iter.getCoord();
         const openvdb::Vec3f voxelIndexSpace = ijk.asVec3d();
@@ -450,6 +511,16 @@ public:
 
     inline bool initialized() const { return mFilter1.initialized() && mFilter2.initialized(); }
 
+    inline index::State state() const
+    {
+        return this->computeState(mFilter1.state(), mFilter2.state());
+    }
+    template <typename LeafT>
+    inline index::State state(const LeafT& leaf) const
+    {
+        return this->computeState(mFilter1.state(leaf), mFilter2.state(leaf));
+    }
+
     template <typename LeafT>
     void reset(const LeafT& leaf) {
         mFilter1.reset(leaf);
@@ -463,6 +534,19 @@ public:
     }
 
 private:
+    inline index::State computeState(   index::State state1,
+                                        index::State state2) const
+    {
+        if (And) {
+            if (state1 == index::NONE || state2 == index::NONE)       return index::NONE;
+            else if (state1 == index::ALL && state2 == index::ALL)    return index::ALL;
+        } else {
+            if (state1 == index::NONE && state2 == index::NONE)       return index::NONE;
+            else if (state1 == index::ALL && state2 == index::ALL)    return index::ALL;
+        }
+        return index::PARTIAL;
+    }
+
     T1 mFilter1;
     T2 mFilter2;
 }; // class BinaryFilter
@@ -498,7 +582,3 @@ struct FilterTraits<BinaryFilter<T0, T1, And>> {
 } // namespace openvdb
 
 #endif // OPENVDB_POINTS_INDEX_FILTER_HAS_BEEN_INCLUDED
-
-// Copyright (c) 2012-2018 DreamWorks Animation LLC
-// All rights reserved. This software is distributed under the
-// Mozilla Public License 2.0 ( http://www.mozilla.org/MPL/2.0/ )
